@@ -17,6 +17,7 @@
 #include "ngx_http_lua_string.h"
 #include "ngx_http_lua_misc.h"
 #include "ngx_http_lua_consts.h"
+#include "ngx_http_lua_shdict.h"
 
 
 static void ngx_http_lua_inject_arg_api(lua_State *L,
@@ -35,6 +36,9 @@ ngx_http_lua_set_by_chunk(lua_State *L, ngx_http_request_t *r, ngx_str_t *val,
     u_char          *err_msg;
     size_t           rlen;
     u_char          *rdata;
+#if (NGX_PCRE)
+    ngx_pool_t      *old_pool;
+#endif
 
     /*  set Lua VM panic handler */
     lua_atpanic(L, ngx_http_lua_atpanic);
@@ -49,7 +53,7 @@ ngx_http_lua_set_by_chunk(lua_State *L, ngx_http_request_t *r, ngx_str_t *val,
 
 #if (NGX_PCRE)
     /* XXX: work-around to nginx regex subsystem */
-    ngx_http_lua_pcre_malloc_init(r->pool);
+    old_pool = ngx_http_lua_pcre_malloc_init(r->pool);
 #endif
 
     /*  protected call user code */
@@ -57,7 +61,7 @@ ngx_http_lua_set_by_chunk(lua_State *L, ngx_http_request_t *r, ngx_str_t *val,
 
 #if (NGX_PCRE)
     /* XXX: work-around to nginx regex subsystem */
-    ngx_http_lua_pcre_malloc_done();
+    ngx_http_lua_pcre_malloc_done(old_pool);
 #endif
 
     if (rc != 0) {
@@ -164,6 +168,10 @@ static void
 ngx_http_lua_set_by_lua_env(lua_State *L, ngx_http_request_t *r, size_t nargs,
         ngx_http_variable_value_t *args)
 {
+    ngx_http_lua_main_conf_t    *lmcf;
+
+    lmcf = ngx_http_get_module_main_conf(r, ngx_http_lua_module);
+
     /*  set nginx request pointer to current lua thread's globals table */
     lua_pushlightuserdata(L, r);
     lua_setglobal(L, GLOBALS_SYMBOL_REQUEST);
@@ -196,11 +204,12 @@ ngx_http_lua_set_by_lua_env(lua_State *L, ngx_http_request_t *r, size_t nargs,
     ngx_http_lua_inject_time_api(L);
     ngx_http_lua_inject_string_api(L);
     ngx_http_lua_inject_variable_api(L);
-    ngx_http_lua_inject_req_api(L);
+    ngx_http_lua_inject_req_api_no_io(L);
     ngx_http_lua_inject_arg_api(L, nargs, args);
 #if (NGX_PCRE)
     ngx_http_lua_inject_regex_api(L);
 #endif
+    ngx_http_lua_inject_shdict_api(lmcf, L);
     ngx_http_lua_inject_misc_api(L);
 
     lua_setfield(L, -2, "ngx");
