@@ -5,9 +5,14 @@
 
 #include "ngx_http_lua_subrequest.h"
 #include "ngx_http_lua_util.h"
+#include "ngx_http_lua_ctx.h"
 #include "ngx_http_lua_contentby.h"
 
 
+#define NGX_HTTP_LUA_SHARE_ALL_VARS     0x01
+#define NGX_HTTP_LUA_COPY_ALL_VARS      0x02
+
+
 #define ngx_http_lua_method_name(m) { sizeof(m) - 1, (u_char *) m " " }
 
 static ngx_str_t  ngx_http_lua_get_method = ngx_http_lua_method_name("GET");
@@ -23,12 +28,16 @@ static ngx_str_t  ngx_http_lua_content_length_header_key =
 
 
 static ngx_int_t ngx_http_lua_set_content_length_header(ngx_http_request_t *r,
-        off_t len);
+    off_t len);
 static ngx_int_t ngx_http_lua_adjust_subrequest(ngx_http_request_t *sr,
-        ngx_uint_t method, ngx_http_request_body_t *body,
-        unsigned share_all_vars);
+    ngx_uint_t method, ngx_http_request_body_t *body, unsigned vars_action,
+    ngx_array_t *extra_vars);
 static int ngx_http_lua_ngx_location_capture(lua_State *L);
 static int ngx_http_lua_ngx_location_capture_multi(lua_State *L);
+static void ngx_http_lua_process_vars_option(ngx_http_request_t *r,
+    lua_State *L, int table, ngx_array_t **varsp);
+static ngx_int_t ngx_http_lua_subrequest_add_extra_vars(ngx_http_request_t *r,
+    ngx_array_t *extra_vars);
 
 
 /* ngx.location.capture is just a thin wrapper around
@@ -70,6 +79,7 @@ ngx_http_lua_ngx_location_capture_multi(lua_State *L)
     ngx_http_post_subrequest_t      *psr;
     ngx_http_lua_ctx_t              *sr_ctx;
     ngx_http_lua_ctx_t              *ctx;
+    ngx_array_t                     *extra_vars;
     ngx_str_t                        uri;
     ngx_str_t                        args;
     ngx_str_t                        extra_args;
@@ -84,12 +94,13 @@ ngx_http_lua_ngx_location_capture_multi(lua_State *L)
     ngx_http_request_body_t         *body;
     int                              type;
     ngx_buf_t                       *b;
-    unsigned                         share_all_vars;
+    unsigned                         vars_action;
     ngx_uint_t                       nsubreqs;
     ngx_uint_t                       index;
     size_t                           sr_statuses_len;
     size_t                           sr_headers_len;
     size_t                           sr_bodies_len;
+    unsigned                         custom_ctx;
 
     n = lua_gettop(L);
     if (n != 1) {
@@ -143,6 +154,8 @@ ngx_http_lua_ngx_location_capture_multi(lua_State *L)
     ctx->done = 0;
     ctx->waiting = 0;
 
+    extra_vars = NULL;
+
     for (index = 0; index < nsubreqs; index++) {
         ctx->waiting++;
 
@@ -151,14 +164,14 @@ ngx_http_lua_ngx_location_capture_multi(lua_State *L)
             return luaL_error(L, "only array-like tables are allowed");
         }
 
+        dd("queries query: top %d", lua_gettop(L));
+
         if (lua_type(L, -1) != LUA_TTABLE) {
             return luaL_error(L, "the query argument %d is not a table, "
                     "but a %s",
                     index, lua_typename(L, lua_type(L, -1)));
         }
 
-        dd("lua top so far: %d", lua_gettop(L));
-
         nargs = lua_objlen(L, -1);
 
         if (nargs != 1 && nargs != 2) {
@@ -168,6 +181,8 @@ ngx_http_lua_ngx_location_capture_multi(lua_State *L)
 
         lua_rawgeti(L, 2, 1); /* queries query uri */
 
+        dd("queries query uri: %d", lua_gettop(L));
+
         dd("first arg in first query: %s", lua_typename(L, lua_type(L, -1)));
 
         body = NULL;
@@ -175,18 +190,30 @@ ngx_http_lua_ngx_location_capture_multi(lua_State *L)
         extra_args.data = NULL;
         extra_args.len = 0;
 
-        share_all_vars = 0;
+        if (extra_vars != NULL) {
+            /* flush out existing elements in the array */
+            extra_vars->nelts = 0;
+        }
+
+        vars_action = 0;
+
+        custom_ctx = 0;
 
         if (nargs == 2) {
             /* check out the options table */
 
-            lua_rawgeti(L, 2, 2);
+            lua_rawgeti(L, 2, 2); /* queries query uri opts */
+
+            dd("queries query uri opts: %d", lua_gettop(L));
 
             if (lua_type(L, 4) != LUA_TTABLE) {
                 return luaL_error(L, "expecting table as the 2nd argument for "
-                        "subrequest %d", index);
+                        "subrequest %d, but got %s", index,
+                        luaL_typename(L, 4));
             }
 
+            dd("queries query uri opts: %d", lua_gettop(L));
+
             /* check the args option */
 
             lua_getfield(L, 4, "args");
@@ -215,26 +242,78 @@ ngx_http_lua_ngx_location_capture_multi(lua_State *L)
 
             lua_pop(L, 1);
 
-            /* check the share_all_vars option */
+            dd("queries query uri opts: %d", lua_gettop(L));
 
-            lua_getfield(L, 4, "share_all_vars");
+            /* check the vars option */
 
-            type = lua_type(L, -1);
+            lua_getfield(L, 4, "vars");
 
-            if (type == LUA_TNIL) {
+            switch (lua_type(L, -1)) {
+            case LUA_TTABLE:
+                ngx_http_lua_process_vars_option(r, L, -1, &extra_vars);
+
+                dd("post process vars top: %d", lua_gettop(L));
+                break;
+
+            case LUA_TNIL:
                 /* do nothing */
+                break;
 
-            } else {
-                if (type != LUA_TBOOLEAN) {
-                    return luaL_error(L, "Bad share_all_vars option value");
-                }
-
-                share_all_vars = lua_toboolean(L, -1);
+            default:
+                return luaL_error(L, "Bad vars option value");
             }
 
             lua_pop(L, 1);
 
-            /* check the method option */
+            dd("queries query uri opts: %d", lua_gettop(L));
+
+            /* check the share_all_vars option */
+
+            lua_getfield(L, 4, "share_all_vars");
+
+            switch (lua_type(L, -1)) {
+            case LUA_TNIL:
+                /* do nothing */
+                break;
+
+            case LUA_TBOOLEAN:
+                if (lua_toboolean(L, -1)) {
+                    vars_action |= NGX_HTTP_LUA_SHARE_ALL_VARS;
+                }
+                break;
+
+            default:
+                return luaL_error(L, "Bad share_all_vars option value");
+            }
+
+            lua_pop(L, 1);
+
+            dd("queries query uri opts: %d", lua_gettop(L));
+
+            /* check the copy_all_vars option */
+
+            lua_getfield(L, 4, "copy_all_vars");
+
+            switch (lua_type(L, -1)) {
+            case LUA_TNIL:
+                /* do nothing */
+                break;
+
+            case LUA_TBOOLEAN:
+                if (lua_toboolean(L, -1)) {
+                    vars_action |= NGX_HTTP_LUA_COPY_ALL_VARS;
+                }
+                break;
+
+            default:
+                return luaL_error(L, "Bad copy_all_vars option value");
+            }
+
+            lua_pop(L, 1);
+
+            dd("queries query uri opts: %d", lua_gettop(L));
+
+            /* check the "method" option */
 
             lua_getfield(L, 4, "method");
 
@@ -253,7 +332,29 @@ ngx_http_lua_ngx_location_capture_multi(lua_State *L)
 
             lua_pop(L, 1);
 
-            /* check the body option */
+            dd("queries query uri opts: %d", lua_gettop(L));
+
+            /* check the "ctx" option */
+
+            lua_getfield(L, 4, "ctx");
+
+            type = lua_type(L, -1);
+
+            if (type != LUA_TNIL) {
+                if (type != LUA_TTABLE) {
+                    return luaL_error(L, "Bad ctx option value type %s, "
+                            "expected a Lua table", lua_typename(L, type));
+                }
+
+                custom_ctx = 1;
+
+            } else {
+                lua_pop(L, 1);
+            }
+
+            dd("queries query uri opts ctx?: %d", lua_gettop(L));
+
+            /* check the "body" option */
 
             lua_getfield(L, 4, "body");
 
@@ -295,11 +396,22 @@ ngx_http_lua_ngx_location_capture_multi(lua_State *L)
                 }
             }
 
-            lua_pop(L, 2); /* pop body and opts table */
+            lua_pop(L, 1); /* pop the body */
+
+            /* stack: queries query uri opts ctx? */
+
+            lua_remove(L, 4);
+
+            /* stack: queries query uri ctx? */
+
+            dd("queries query uri ctx?: %d", lua_gettop(L));
+
         } else {
             method = NGX_HTTP_GET;
         }
 
+        /* stack: queries query uri ctx? */
+
         n = lua_gettop(L);
         dd("top size so far: %d", n);
 
@@ -379,14 +491,31 @@ ngx_http_lua_ngx_location_capture_multi(lua_State *L)
 
         ngx_http_set_ctx(sr, sr_ctx, ngx_http_lua_module);
 
-        rc = ngx_http_lua_adjust_subrequest(sr, method, body, share_all_vars);
+        rc = ngx_http_lua_adjust_subrequest(sr, method, body, vars_action,
+                extra_vars);
 
         if (rc != NGX_OK) {
             return luaL_error(L, "failed to adjust the subrequest: %d",
                     (int) rc);
         }
 
-        lua_pop(L, 2); /* pop the subrequest argument and uri */
+        dd("queries query uri opts ctx? %d", lua_gettop(L));
+
+        /* stack: queries query uri ctx? */
+
+        if (custom_ctx) {
+            ngx_http_lua_ngx_set_ctx_helper(L, sr, sr_ctx, -1);
+            lua_pop(L, 3);
+
+        } else {
+            lua_pop(L, 2);
+        }
+
+        /* stack: queries */
+    }
+
+    if (extra_vars) {
+        ngx_array_destroy(extra_vars);
     }
 
     return lua_yield(L, 0);
@@ -395,11 +524,13 @@ ngx_http_lua_ngx_location_capture_multi(lua_State *L)
 
 static ngx_int_t
 ngx_http_lua_adjust_subrequest(ngx_http_request_t *sr, ngx_uint_t method,
-        ngx_http_request_body_t *body, unsigned share_all_vars)
+    ngx_http_request_body_t *body, unsigned vars_action,
+    ngx_array_t *extra_vars)
 {
     ngx_http_request_t          *r;
     ngx_int_t                    rc;
     ngx_http_core_main_conf_t   *cmcf;
+    size_t                       size;
 
     r = sr->parent;
 
@@ -458,22 +589,198 @@ ngx_http_lua_adjust_subrequest(ngx_http_request_t *sr, ngx_uint_t method,
         sr->headers_in.headers.last = &sr->headers_in.headers.part;
     }
 
-    if (! share_all_vars) {
+    if (!(vars_action & NGX_HTTP_LUA_SHARE_ALL_VARS)) {
         /* we do not inherit the parent request's variables */
         cmcf = ngx_http_get_module_main_conf(sr, ngx_http_core_module);
 
-        sr->variables = ngx_pcalloc(sr->pool, cmcf->variables.nelts
-                                * sizeof(ngx_http_variable_value_t));
+        size = cmcf->variables.nelts * sizeof(ngx_http_variable_value_t);
 
-        if (sr->variables == NULL) {
-            return NGX_HTTP_INTERNAL_SERVER_ERROR;
+        if (vars_action & NGX_HTTP_LUA_COPY_ALL_VARS) {
+
+            sr->variables = ngx_palloc(sr->pool, size);
+            if (sr->variables == NULL) {
+                return NGX_ERROR;
+            }
+
+            ngx_memcpy(sr->variables, r->variables, size);
+
+        } else {
+
+            /* we do not inherit the parent request's variables */
+
+            sr->variables = ngx_pcalloc(sr->pool, size);
+            if (sr->variables == NULL) {
+                return NGX_ERROR;
+            }
         }
     }
 
+    return ngx_http_lua_subrequest_add_extra_vars(sr, extra_vars);
+}
+
+
+static ngx_int_t
+ngx_http_lua_subrequest_add_extra_vars(ngx_http_request_t *sr,
+       ngx_array_t *extra_vars)
+{
+    ngx_http_core_main_conf_t   *cmcf;
+    ngx_http_variable_t         *v;
+    ngx_http_variable_value_t   *vv;
+    u_char                      *val;
+    u_char                      *p;
+    ngx_uint_t                   i, hash;
+    ngx_str_t                    name;
+    size_t                       len;
+    ngx_hash_t                  *variables_hash;
+    ngx_keyval_t                *var;
+
+    /* set any extra variables that were passed to the subrequest */
+
+    if (extra_vars == NULL || extra_vars->nelts == 0) {
+        return NGX_OK;
+    }
+
+    cmcf = ngx_http_get_module_main_conf(sr, ngx_http_core_module);
+
+    variables_hash = &cmcf->variables_hash;
+
+    var = extra_vars->elts;
+
+    for (i = 0; i < extra_vars->nelts; i++, var++) {
+        /* copy the variable's name and value because they are allocated
+         * by the lua VM */
+
+        len = var->key.len + var->value.len;
+
+        p = ngx_pnalloc(sr->pool, len);
+        if (p == NULL) {
+            return NGX_ERROR;
+        }
+
+        name.data = p;
+        name.len = var->key.len;
+
+        p = ngx_copy(p, var->key.data, var->key.len);
+
+        hash = ngx_hash_strlow(name.data, name.data, name.len);
+
+        val = p;
+        len = var->value.len;
+
+        ngx_memcpy(p, var->value.data, len);
+
+        v = ngx_hash_find(variables_hash, hash, name.data, name.len);
+
+        if (v) {
+            if (!(v->flags & NGX_HTTP_VAR_CHANGEABLE)) {
+                ngx_log_error(NGX_LOG_ERR, sr->connection->log, 0,
+                              "variable \"%V\" not changeable", &name);
+                return NGX_HTTP_INTERNAL_SERVER_ERROR;
+            }
+
+            if (v->set_handler) {
+                vv = ngx_palloc(sr->pool, sizeof(ngx_http_variable_value_t));
+                if (vv == NULL) {
+                    return NGX_ERROR;
+                }
+
+                vv->valid = 1;
+                vv->not_found = 0;
+                vv->no_cacheable = 0;
+
+                vv->data = val;
+                vv->len = len;
+
+                v->set_handler(sr, vv, v->data);
+
+                ngx_log_debug2(NGX_LOG_DEBUG_HTTP, sr->connection->log, 0,
+                      "variable \"%V\" set to value \"%v\"", &name, vv);
+
+                continue;
+            }
+
+            if (v->flags & NGX_HTTP_VAR_INDEXED) {
+                vv = &sr->variables[v->index];
+
+                vv->valid = 1;
+                vv->not_found = 0;
+                vv->no_cacheable = 0;
+
+                vv->data = val;
+                vv->len = len;
+
+                ngx_log_debug2(NGX_LOG_DEBUG_HTTP, sr->connection->log, 0,
+                      "variable \"%V\" set to value \"%v\"", &name, vv);
+
+                continue;
+            }
+        }
+
+        ngx_log_error(NGX_LOG_ERR, sr->connection->log, 0,
+                      "variable \"%V\" cannot be assigned a value (maybe you "
+                      "forgot to define it first?) ", &name);
+
+        return NGX_ERROR;
+    }
+
     return NGX_OK;
 }
 
 
+static void
+ngx_http_lua_process_vars_option(ngx_http_request_t *r, lua_State *L,
+        int table, ngx_array_t **varsp)
+{
+    ngx_array_t         *vars;
+    ngx_keyval_t        *var;
+
+    if (table < 0) {
+        table = lua_gettop(L) + table + 1;
+    }
+
+    vars = *varsp;
+
+    if (vars == NULL) {
+
+        vars = ngx_array_create(r->pool, 4, sizeof(ngx_keyval_t));
+        if (vars == NULL) {
+            dd("here");
+            luaL_error(L, "out of memory");
+            return;
+        }
+
+        *varsp = vars;
+    }
+
+    lua_pushnil(L);
+    while (lua_next(L, table) != 0) {
+
+        if (lua_type(L, -2) != LUA_TSTRING) {
+            luaL_error(L, "attempt to use a non-string key in the "
+                    "\"vars\" option table");
+            return;
+        }
+
+        if (!lua_isstring(L, -1)) {
+            luaL_error(L, "attempt to use bad variable value type %s",
+                       luaL_typename(L, -1));
+        }
+
+        var = ngx_array_push(vars);
+        if (var == NULL) {
+            dd("here");
+            luaL_error(L, "out of memory");
+            return;
+        }
+
+        var->key.data = (u_char *) lua_tolstring(L, -2, &var->key.len);
+        var->value.data = (u_char *) lua_tolstring(L, -1, &var->value.len);
+
+        lua_pop(L, 1);
+    }
+}
+
+
 ngx_int_t
 ngx_http_lua_post_subrequest(ngx_http_request_t *r, void *data, ngx_int_t rc)
 {
@@ -669,7 +976,16 @@ ngx_http_lua_set_content_length_header(ngx_http_request_t *r, off_t len)
 
     h->value.len = ngx_sprintf(h->value.data, "%O", len) - h->value.data;
 
-    h->hash = 1;
+    h->hash = ngx_hash(ngx_hash(ngx_hash(ngx_hash(ngx_hash(ngx_hash(ngx_hash(
+            ngx_hash(ngx_hash(ngx_hash(ngx_hash(ngx_hash(
+            ngx_hash('c', 'o'), 'n'), 't'), 'e'), 'n'), 't'), '-'), 'l'), 'e'),
+            'n'), 'g'), 't'), 'h');
+
+#if 0
+    dd("content length hash: %lu == %lu", (unsigned long) h->hash,
+            ngx_hash_key_lc((u_char *) "Content-Length",
+            sizeof("Content-Length") - 1));
+#endif
 
     dd("r content length: %.*s",
             (int)r->headers_in.content_length->value.len,
@@ -734,6 +1050,9 @@ ngx_http_lua_handle_subreq_responses(ngx_http_request_t *r,
 
     u_char                  buf[sizeof("Mon, 28 Sep 1970 06:00:00 GMT") - 1];
 
+    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
+            "lua handle subrequest responses");
+
     for (index = 0; index < ctx->nsubreqs; index++) {
         dd("summary: reqs %d, subquery %d, waiting %d, req %.*s",
                 (int) ctx->nsubreqs,
